package mockingjay

import (
	"bytes"
	"encoding/json"
	"net/http"
	"net/http/httptest"
	"reflect"
	"strings"
	"testing"
)

var (
	cdcDisabled = false
)

const cannedResponse = "world"
const cannedResponse2 = "worldy world world"
const testURL = "/hello"
const testURL2 = "/hello2"
const testEndpointName = "Test 123"

func TestItReturnsCannedResponses(t *testing.T) {
	responseHeaders := make(map[string]string)
	responseHeaders["Content-Type"] = "application/json"

	endpoint := FakeEndpoint{
		Name:        "Fake 1",
		CDCDisabled: cdcDisabled,
		Request: Request{
			URI:     testURL,
			Method:  "GET",
			Headers: nil,
			Body:    "",
		},
		Response: response{
			Code:    http.StatusCreated,
			Body:    cannedResponse,
			Headers: responseHeaders,
		},
	}
	secondEndpoint := FakeEndpoint{
		Name:        "Fake 2",
		CDCDisabled: cdcDisabled,
		Request: Request{
			URI:     testURL2,
			Method:  "GET",
			Headers: nil,
			Body:    "",
		},
		Response: response{
			Code:    http.StatusCreated,
			Body:    cannedResponse2,
			Headers: nil,
		},
	}

	server := NewServer([]FakeEndpoint{endpoint, secondEndpoint})

	request, _ := http.NewRequest("GET", testURL, nil)
	responseReader := httptest.NewRecorder()

	server.ServeHTTP(responseReader, request)

	if responseReader.Code != http.StatusCreated {
		t.Fatal("Expected a 201 (status created) but got", responseReader.Code)
	}

	if responseReader.Body.String() != cannedResponse {
		t.Errorf("Expected canned response to be returned, got [%s] ", responseReader.Body.String())
	}

	if responseReader.Header().Get("Content-Type") != "application/json" {
		t.Errorf("Expected response header to be set")
	}

	responseReader = httptest.NewRecorder()
	requestTwo, _ := http.NewRequest("GET", testURL2, nil)

	server.ServeHTTP(responseReader, requestTwo)

	if responseReader.Body.String() != cannedResponse2 {
		t.Errorf("Expected second endpoint's canned response, got %s", responseReader.Body.String())
	}
}

func TestItCanCreateEndpoints(t *testing.T) {
	server := NewServer([]FakeEndpoint{})

	newEndpoint := FakeEndpoint{
		Name: "New endpoint",
		Request: Request{
			URI:    "/foo",
			Method: "GET",
			Body:   "Blah blah",
		},
		Response: response{
			Code: 200,
			Body: "SUPER",
		},
	}

	newEndpointJSON, _ := json.Marshal(newEndpoint)

	req, _ := http.NewRequest("POST", newEndpointURL, bytes.NewReader(newEndpointJSON))

	responseReader := httptest.NewRecorder()

	server.ServeHTTP(responseReader, req)

	if responseReader.Code != http.StatusCreated {
		t.Fatal("Didnt create new endpoint", responseReader.Code, responseReader.Body.String())
	}

	if len(server.endpoints) != 1 {
		t.Fatal("The number of endpoints didnt increate after creating")
	}

	if !reflect.DeepEqual(server.endpoints[0], newEndpoint) {
		t.Errorf("New endpoint was not correctly added to the server %v", server.endpoints[0])
	}
}

func TestItReturnsBadRequestWhenMakingABadNewEndpoint(t *testing.T) {
	server := NewServer([]FakeEndpoint{})

	badBody := []byte("blah")
	req, _ := http.NewRequest("POST", newEndpointURL, bytes.NewReader(badBody))
	responseReader := httptest.NewRecorder()

	server.ServeHTTP(responseReader, req)

	if responseReader.Code != http.StatusBadRequest {
		t.Error("Expected to get a bad request but got", responseReader.Code)
	}
}

func TestItReturns404WhenUriIsWrong(t *testing.T) {
	mjReq := Request{
		URI:    testURL,
		Method: "GET"}
	endpoint := FakeEndpoint{testEndpointName, cdcDisabled, mjReq, response{http.StatusCreated, cannedResponse, nil}}
	server := NewServer([]FakeEndpoint{endpoint})
	requestBody := "some body"

	request, _ := http.NewRequest("GET", "/bums", strings.NewReader(requestBody))
	responseReader := httptest.NewRecorder()

	server.ServeHTTP(responseReader, request)

	if responseReader.Code != http.StatusNotFound {
		t.Error("Expected to get a 404")
	}

	if !strings.Contains(responseReader.Body.String(), requestBody) {
		t.Errorf("Expected request body to be returned %v", responseReader.Body.String())
	}
}

func TestItReturns404WhenMethodIsWrong(t *testing.T) {
	mjReq := Request{URI: testURL, Method: "GET"}
	endpoint := FakeEndpoint{testEndpointName, cdcDisabled, mjReq, response{http.StatusCreated, cannedResponse, nil}}
	server := NewServer([]FakeEndpoint{endpoint})

	request, _ := http.NewRequest("POST", "/hello", nil)
	request.Header.Set("content-type", "application/bob")

	responseReader := httptest.NewRecorder()

	server.ServeHTTP(responseReader, request)

	if responseReader.Code != http.StatusNotFound {
		t.Error("Expected to get a 404")
	}

	if !strings.Contains(responseReader.Body.String(), "application/bob") {
		t.Log(responseReader.Body.String())
		t.Error("Request headers were not added to the response info")
	}
}

func TestItDoesContentNegotiation(t *testing.T) {
	contentTypes := make(map[string]string)
	contentTypes["Content-Type"] = "application/json"

	mjReq := Request{URI: testURL, Method: "GET", Headers: contentTypes}
	endpoint := FakeEndpoint{testEndpointName, cdcDisabled, mjReq, response{http.StatusCreated, cannedResponse, nil}}
	server := NewServer([]FakeEndpoint{endpoint})

	requestWithIncorrectHeaderValue, _ := http.NewRequest("GET", testURL, nil)
	requestWithIncorrectHeaderValue.Header.Add("Content-Type", "application/xml")
	responseReader := httptest.NewRecorder()

	server.ServeHTTP(responseReader, requestWithIncorrectHeaderValue)

	if responseReader.Code != http.StatusNotFound {
		t.Error("Expected to get a 404 because we didnt set a content type header when it was expected")
	}

	requestWithNoHeaderAtAll, _ := http.NewRequest("GET", testURL, nil)
	responseReader = httptest.NewRecorder()

	server.ServeHTTP(responseReader, requestWithNoHeaderAtAll)

	if responseReader.Code != http.StatusNotFound {
		t.Error("Expected to get a 404 because we didnt set a content type header when it was expected")
	}

	requestWithDifferentCasedHeader, _ := http.NewRequest("GET", testURL, nil)
	requestWithDifferentCasedHeader.Header.Add("Content-TYPE", "application/json")
	responseReader = httptest.NewRecorder()

	server.ServeHTTP(responseReader, requestWithDifferentCasedHeader)

	if responseReader.Code != http.StatusCreated {
		t.Errorf("Expected request to match even though the header name was differently cased : \n%d", responseReader.Code)
	}
}

func TestItSendsRequestBodies(t *testing.T) {
	body := "some body"
	expectedStatus := http.StatusInternalServerError

	mjReq := Request{URI: testURL, Method: "POST", Body: body}
	endpoint := FakeEndpoint{testEndpointName, cdcDisabled, mjReq, response{expectedStatus, "", nil}}
	server := NewServer([]FakeEndpoint{endpoint})

	requestWithoutBody, _ := http.NewRequest("POST", testURL, nil)
	responseReader := httptest.NewRecorder()

	server.ServeHTTP(responseReader, requestWithoutBody)

	if responseReader.Code != http.StatusNotFound {
		t.Error("Expected to get a 404 because we didnt set a request body when it was expected")
	}

	requestWithBody, _ := http.NewRequest("POST", testURL, strings.NewReader(body))
	responseReader = httptest.NewRecorder()

	server.ServeHTTP(responseReader, requestWithBody)

	if responseReader.Code != expectedStatus {
		t.Error("Expected request to succeed but it didnt")
	}
}

func TestItRecordsIncomingRequests(t *testing.T) {
	wildcardBody := "*"
	expectedStatus := http.StatusOK

	mjReq := Request{URI: testURL, Method: "POST", Body: wildcardBody}
	config := FakeEndpoint{testEndpointName, cdcDisabled, mjReq, response{expectedStatus, "", nil}}
	server := NewServer([]FakeEndpoint{config})

	requestWithDifferentBody, _ := http.NewRequest("POST", testURL, strings.NewReader("This body isnt what we said but it should match"))
	responseReader := httptest.NewRecorder()

	server.ServeHTTP(responseReader, requestWithDifferentBody)

	if len(server.requests) != 1 {
		t.Fatalf("Expected one request to be recorded but got %d", len(server.requests))
	}

	if server.requests[0].Method != "POST" {
		t.Error("It doesnt look like it recorded the request properly")
	}
}

func TestItRespectsURLencoding(t *testing.T) {
	escapedURL := "/document/10.1007%2Fs00414-006-0114-x"
	body := "some body"
	expectedStatus := http.StatusOK

	mjReq := Request{URI: escapedURL, Method: "POST", Body: body}
	config := FakeEndpoint{testEndpointName, cdcDisabled, mjReq, response{expectedStatus, "", nil}}
	server := NewServer([]FakeEndpoint{config})

	request, _ := http.NewRequest("POST", escapedURL, strings.NewReader(body))
	responseReader := httptest.NewRecorder()

	server.ServeHTTP(responseReader, request)

	if responseReader.Code != expectedStatus {
		t.Errorf("It looks like it didnt respect the escaped url, got a %d", responseReader.Code)
	}
}

func TestItReturnsListOfEndpoints(t *testing.T) {
	mjReq := Request{URI: testURL, Method: "GET"}
	endpoint := FakeEndpoint{testEndpointName, cdcDisabled, mjReq, response{http.StatusCreated, cannedResponse, nil}}
	server := NewServer([]FakeEndpoint{endpoint})

	request, _ := http.NewRequest("GET", endpointsURL, nil)
	responseReader := httptest.NewRecorder()

	server.ServeHTTP(responseReader, request)

	if responseReader.Code != http.StatusOK {
		t.Error("Didnt get a 200 from the endpoints url, got", responseReader.Code)
	}

	if responseReader.HeaderMap["Content-Type"][0] != "application/json" {
		t.Error("Expected content type header to be set to json")
	}

	var endpointResponse []FakeEndpoint
	err := json.Unmarshal(responseReader.Body.Bytes(), &endpointResponse)

	if err != nil {
		t.Fatal("Unable to parse JSON from endpoints URL", err)
	}

	if !reflect.DeepEqual(endpointResponse[0], endpoint) {
		t.Error("The endpoint returned doesnt match what the server was set up with, got", endpointResponse, "expected", endpoint)
	}
}
